@@ -1,8 +1,8 @@
 /* -*- C -*-
    +----------------------------------------------------------------------+
-   | PHP Version 5                                                        |
+   | PHP Version 7                                                        |
    +----------------------------------------------------------------------+
-   | Copyright (c) 1997-2016 The PHP Group                                |
+   | Copyright (c) 1997-2018 The PHP Group                                |
    +----------------------------------------------------------------------+
    | This source file is subject to version 3.01 of the PHP license,      |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,13 +12,11 @@
    | obtain it through the world-wide-web, please send a note to          |
    | license@php.net so we can mail you a copy immediately.               |
    +----------------------------------------------------------------------+
-   | Authors: Andi Gutmans <andi@zend.com>                                |
-   |          Zeev Suraski <zeev@zend.com>                                |
+   | Authors: Andi Gutmans <andi@php.net>                                 |
+   |          Zeev Suraski <zeev@php.net>                                 |
    +----------------------------------------------------------------------+
  */
 
-/* $Id$ */
-
 #include "php.h"
 #include "php_main.h"
 #include "zend_modules.h"
@@ -28,7 +26,6 @@
 #include <stdio.h>
 
 #include "ext/date/php_date.h"
-#include "ext/ereg/php_ereg.h"
 #include "ext/libxml/php_libxml.h"
 #include "ext/openssl/php_openssl.h"
 #include "ext/pcre/php_pcre.h"
@@ -45,9 +42,8 @@
 #include "ext/mysqlnd/php_mysqlnd.h"
 
 
-static zend_module_entry *php_builtin_extensions[] = {
+static zend_module_entry * const php_builtin_extensions[] = {
 	phpext_date_ptr,
-	phpext_ereg_ptr,
 	phpext_libxml_ptr,
 	phpext_openssl_ptr,
 	phpext_pcre_ptr,
@@ -67,9 +63,9 @@ static zend_module_entry *php_builtin_extensions[] = {
 
 #define EXTCOUNT (sizeof(php_builtin_extensions)/sizeof(zend_module_entry *))
 
-PHPAPI int php_register_internal_extensions(TSRMLS_D)
+PHPAPI int php_register_internal_extensions(void)
 {
-	return php_register_extensions(php_builtin_extensions, EXTCOUNT TSRMLS_CC);
+	return php_register_extensions(php_builtin_extensions, EXTCOUNT);
 }
 
 /*
